# -*- coding:Utf-8 -*-
from __future__ import unicode_literals

from pybtex.style.names import BaseNameStyle, name_part
from pybtex.style.template import join


class FirstLastStyle(BaseNameStyle):

    def format(self, person, abbr=False):
        r"""
        Format names similarly to {vv~}{ll}{, jj}{, f.} in BibTeX.

        >>> from pybtex.database import Person
        >>> name = Person(string=r"Charles Louis Xavier Joseph de la Vall{\'e}e Poussin")
        >>> firstlast = NameStyle().format

        >>> print(firstlast(name).format().render_as('latex'))
        Charles Louis Xavier~Joseph de~la Vall{é}e~Poussin
        >>> print(firstlast(name).format().render_as('html'))
        Charles Louis Xavier&nbsp;Joseph de&nbsp;la Vall<span class="bibtex-protected">é</span>e&nbsp;Poussin

        >>> print(firstlast(name, abbr=True).format().render_as('latex'))
        C.~L. X.~J. de~la Vall{é}e~Poussin
        >>> print(firstlast(name, abbr=True).format().render_as('html'))
        C.&nbsp;L. X.&nbsp;J. de&nbsp;la Vall<span class="bibtex-protected">é</span>e&nbsp;Poussin

        >>> name = Person(first='First', last='Last', middle='Middle')
        >>> print(firstlast(name).format().render_as('latex'))
        First~Middle Last
        >>> print(firstlast(name, abbr=True).format().render_as('latex'))
        F.~M. Last

        """
        return join[
            name_part(abbr=abbr)[person.rich_first_names + person.rich_middle_names],
            name_part(before=' ')[person.rich_prelast_names],
            name_part(before=' ')[person.rich_last_names],
        ]



# -*- coding:Utf-8 -*-
from __future__ import unicode_literals

import re
import six

from pybtex.plugin import find_plugin
from pybtex.style.formatting import BaseStyle, toplevel
from pybtex.style.template import (
    field, first_of, href, join, optional, optional_field, sentence, tag,
    together, words, node, FieldIsMissing
)

from pybtex.richtext import Text, Symbol


firstlast = find_plugin('pybtex.style.names', 'lastfirst')()

if six.PY2:
    def format_pages(text):
        dash_re = re.compile(r'-+')
        pages = Text(Symbol('ndash')).join(
            re.split(dash_re, six.text_type(text)))
        if re.search('[-‒–—―]', six.text_type(text)):
            return Text("pp.", Symbol('nbsp'), pages)
        return Text("p.", Symbol('nbsp'), pages)
else:
    def format_pages(text):
        dash_re = re.compile(r'-+')
        pages = Text(Symbol('ndash')).join(text.split(dash_re))
        if re.search('[-‒–—―]', str(text)):
            return Text("pp.", Symbol('nbsp'), pages)
        return Text("p.", Symbol('nbsp'), pages)


pages = field('pages', apply_func=format_pages)
date = words[field('year'), optional[", ", field('month')]]


@node
def apa_names(children, context, role, **kwargs):
    """
    Returns formatted names as an APA compliant reference list citation.
    """
    assert not children

    try:
        persons = context['entry'].persons[role]
    except KeyError:
        raise FieldIsMissing(role, context['entry'])

    style = context['style']

    if len(persons) > 7:
        persons = persons[:6] + persons[-1:]
        formatted_names = [style.format_name(
            person, style.abbreviate_names) for person in persons]
        return join(sep=', ', last_sep=', … ')[
            formatted_names].format_data(context)
    else:
        formatted_names = [style.format_name(
            person, style.abbreviate_names) for person in persons]
        return join(sep=', ', sep2=', & ', last_sep=', & ')[
            formatted_names].format_data(context)


@node
def editor_names(children, context, with_suffix=True, **kwargs):
    """
    Returns formatted editor names for inbook.
    """
    assert not children

    try:
        editors = context['entry'].persons['editor']
    except KeyError:
        raise FieldIsMissing('editor', context['entry'])

    formatted_names = [
        firstlast.format(editor, True) for editor in editors]

    if with_suffix:
        return words[
            join(sep=', ', sep2=', & ', last_sep=', & ')[formatted_names],
            "(Eds.)" if len(editors) > 1 else "(Ed.)"
        ].format_data(context)

    return join(sep=', ', sep2=', & ', last_sep=', & ')[
        formatted_names
    ].format_data(context)


class APAStyle(BaseStyle):
    name = 'apa'
    default_name_style = 'lastfirst'
    default_sorting_style = 'author_year_title'
    default_label_style = 'apa'

    def __init__(self, *args, **kwargs):
        super(APAStyle, self).__init__(*args, **kwargs)
        self.abbreviate_names = True

    def format_names(self, role, as_sentence=True):
        formatted_names = apa_names(role)
        if as_sentence:
            return sentence(capfirst=False)[formatted_names]
        else:
            return formatted_names

    def format_author_or_editor_and_date(self, e):
        if 'author' in e.persons and 'editor' in e.persons:
            return sentence(sep=' ')[
                self.format_names('author'), join["(", date, ")."],
                self.format_editor(e, as_sentence=False)]
        elif 'author' in e.persons:
            return sentence(sep=' ')[
                self.format_names('author'), join["(", date, ")"]]
        else:
            return sentence(sep=' ')[
                self.format_editor(e, as_sentence=False),
                join["(", date, ")"]]

    def format_editor(self, e, as_sentence=True):
        editors = self.format_names('editor', as_sentence=False)
        if 'editor' not in e.persons:
            # when parsing the template, a FieldIsMissing exception
            # will be thrown anyway; no need to do anything now,
            # just return the template that will throw the exception
            return editors
        if len(e.persons['editor']) > 1:
            word = '(Eds.)'
        else:
            word = '(Ed.)'
        result = join(sep=' ')[editors, word]
        if as_sentence:
            return sentence[result]
        else:
            return result

    def format_volume(self, e, for_article=False):
        prefix = "Vol."
        if for_article:
            return join[
                tag('em')[field('volume')],
                optional['(', field('number'), ')'],
            ]
        else:
            return optional[together[prefix, field('volume')]]

    def format_title(self, e, which_field, as_sentence=True):
        formatted_title = field(
            which_field, apply_func=lambda text: text.capitalize()
        )
        if as_sentence:
            return sentence[formatted_title]
        else:
            return formatted_title

    def format_btitle(self, e, which_field, as_sentence=True):
        formatted_title = tag('em')[field(which_field)]
        if as_sentence:
            return sentence[formatted_title]
        else:
            return formatted_title

    def format_web_refs(self, e):
        # Based on urlbst output.web.refs
        return sentence(add_period=False)[
            optional[self.format_url(e)],
            optional[self.format_eprint(e)],
            optional[self.format_pubmed(e)],
            optional[self.format_doi(e)],
        ]

    def format_url(self, e):
        # Based on urlbst format.url
        return words[
            'URL:',
            href[
                field('url', raw=True),
                field('url', raw=True)
            ]
        ]

    def format_pubmed(self, e):
        # Based on urlbst format.pubmed
        return href[
            join[
                'https://www.ncbi.nlm.nih.gov/pubmed/',
                field('pubmed', raw=True)
            ],
            join[
                'PMID:',
                field('pubmed', raw=True)
            ]
        ]

    def format_doi(self, e):
        # Based on urlbst format.doi
        return href[
            join[
                'https://doi.org/',
                field('doi', raw=True)
            ],
            join[
                'doi:',
                field('doi', raw=True)
            ]
        ]

    def format_eprint(self, e):
        # Based on urlbst format.eprint
        return href[
            join[
                'https://arxiv.org/abs/',
                field('eprint', raw=True)
            ],
            join[
                'arXiv:',
                field('eprint', raw=True)
            ]
        ]

    def get_article_template(self, e):
        # Required fields: author, title, journal, year
        # Optional fields: volume, number, pages, month, note, key
        volume_and_pages = first_of[
            # Volume and pages, with optional issue number
            optional[
                join[
                    self.format_volume(e, for_article=True),
                    optional[', ', field('pages')]
                ],
            ],
            # Pages only
            pages,
        ]
        template = toplevel[
            self.format_names('author'),
            sentence[
                join["(", date, ")"]
            ],
            self.format_title(e, 'title'),
            sentence[
                tag('em')[field('journal')],
                optional[volume_and_pages],
            ],
            sentence[optional_field('note')],
            self.format_web_refs(e),
        ]
        return template

    def get_book_template(self, e):
        # Required fields: author/editor, title, publisher, year
        # Optional fields: volume, series, address, edition, month, note, key,
        #                  isbn
        return toplevel[
            self.format_author_or_editor_and_date(e),
            sentence(sep=' ')[
                self.format_btitle(e, 'title'),
                optional[
                    sentence[
                        optional[field('edition'), ' ed.'],
                        self.format_volume(e),
                    ]
                ]
            ],
            sentence(sep=': ')[
                optional_field('address'),
                field('publisher'),
            ],
            sentence[optional_field('note')],
        ]

    def get_booklet_template(self, e):
        # Required fields: title
        # Optional fields: author, howpublished, address, month, year, note,
        #                  key
        return toplevel[
            sentence(sep=' ')[
                first_of[
                    optional[self.format_names(
                        'author', as_sentence=False)],
                    "None to claim their bones"
                ],
                join["(", first_of[optional[date], "n.d."], ")"]
            ],
            self.format_title(e, 'title'),
            sentence[optional_field('address')],
            sentence[optional_field('note')],
        ]

    def get_inbook_template(self, e):
        # Required fields: author/editor, title, chapter/pages, publisher, year
        # Optional fields: volume, series, address, edition, month, note, key
        return toplevel[
            sentence(sep=' ')[
                self.format_names('author'),
                join["(", date, ")"]
            ],
            self.format_title(e, 'title'),
            sentence(sep=' ')[
                optional[
                    "In ",
                    editor_names(),
                    ","
                ],
                self.format_btitle(e, 'booktitle', as_sentence=False),
                optional[
                    join[
                        "(",
                        sentence(add_period=False)[
                            optional[field('edition'), ' ed.'],
                            self.format_volume(e),
                            pages,
                        ],
                        ")"
                    ]
                ]
            ],
            sentence(sep=': ')[
                optional_field('address'),
                field('publisher'),
            ],
            sentence[optional_field('note')],
        ]

    def get_incollection_template(self, e):
        # Required fields: author, title, booktitle, year
        # Optional fields: editor, pages, organization, publisher, address,
        #                  month, note, key
        return toplevel[
            self.format_author_or_editor_and_date(e),
            self.format_title(e, 'title'),
            sentence(sep=' ')[
                self.format_btitle(e, 'booktitle', as_sentence=False),
                optional["(", pages, ")"]
            ],
            sentence(sep=': ')[
                optional_field('address'),
                optional_field('publisher'),
            ],
            sentence[optional_field('note')],
        ]

    def get_inproceedings_template(self, e):
        # Required fields: author, title, booktitle, year
        # Optional fields: editor, pages, organization, publisher, address,
        #                  month, note, key
        return toplevel[
            self.format_author_or_editor_and_date(e),
            self.format_title(e, 'title'),
            sentence(sep=' ')[
                self.format_btitle(e, 'booktitle', as_sentence=False),
                optional["(", pages, ")"]
            ],
            sentence(sep=': ')[
                optional_field('address'),
                optional_field('publisher'),
            ],
            sentence[optional_field('note')],
            self.format_web_refs(e)
        ]

    def get_manual_template(self, e):
        # Required fields: title
        # Optional fields: author, organization, address, edition, month, year,
        #                  note, key
        return toplevel[
            sentence(sep=' ')[
                first_of[
                    optional[self.format_names('author', as_sentence=False)],
                    optional_field('organization'),
                    "None to claim their bones"
                ],
                join["(", first_of[optional[date], "n.d."], ")"]
            ],
            self.format_btitle(e, 'title'),
            sentence[optional_field('address')],
            sentence[optional_field('note')],
            self.format_web_refs(e)
        ]

    def get_mastersthesis_template(self, e):
        # Required fields: author, title, school, year
        # Optional fields: address, month, note, key
        return toplevel[
            sentence(sep=' ')[
                self.format_names('author'),
                join["(", date, ")"]
            ],
            sentence(sep=' ')[
                self.format_btitle(e, 'title', as_sentence=False),
                "(Master's thesis)"
            ],
            sentence[
                field('school'),
                optional_field('address'),
            ],
            sentence[optional_field('note')],
        ]

    def get_misc_template(self, e):
        # Required fields: aucun
        # Optional fields: author, title, howpublished, month, year, note, key,
        #                  type
        template = toplevel[
            sentence(sep=' ')[
                first_of[
                    optional[self.format_names('author', as_sentence=False)],
                    optional_field('publisher'),
                    "None to claim their bones"
                ],
                join["(", first_of[optional[date], "n.d."], ")"]
            ],
            optional[self.format_title(e, 'title')],
            sentence[optional_field('note')],
            self.format_web_refs(e)
        ]
        return template

    def get_phdthesis_template(self, e):
        # Required fields: author, title, school, year
        # Optional fields: address, month, note, key
        return toplevel[
            sentence(sep=' ')[
                self.format_names('author'),
                join["(", date, ")"]
            ],
            sentence(sep=' ')[
                self.format_btitle(e, 'title', as_sentence=False),
                "(Doctoral dissertation)"
            ],
            sentence[
                field('school'),
                optional_field('address'),
            ],
            sentence[optional_field('note')],
        ]

    def get_proceedings_template(self, e):
        # Required fields: title, year
        # Optional fields: editor, publisher, organization, address, month,
        #                  note, key
        return toplevel[
            sentence(sep=' ')[
                first_of[
                    optional[self.format_editor(e, as_sentence=False)],
                    optional_field('organization'),
                    "None to claim their bones"
                ],
                join["(", first_of[optional[date], "n.d."], ")"]
            ],
            self.format_btitle(e, 'title'),
            sentence(sep=': ')[
                optional_field('address'),
                optional_field('publisher'),
            ],
            sentence[optional_field('note')],
        ]

    def get_techreport_template(self, e):
        # Required fields: author, title, institution, year
        # Optional fields: type, number, address, month, note, key
        return toplevel[
            sentence(sep=' ')[
                self.format_names('author', as_sentence=False),
                join["(", date, ")"]
            ],
            self.format_btitle(e, 'title'),
            sentence[field('institution')],
            sentence[optional_field('note')],
        ]

    def get_unpublished_template(self, e):
        # Required fields: author, title, note
        # Optional fields: month, year, key
        template = toplevel[
            sentence(sep=' ')[
                self.format_names('author', as_sentence=False),
                join["(", first_of[optional[date], "n.d."], ")"]
            ],
            self.format_btitle(e, 'title'),
            sentence[field('note')],
        ]
        return template



import os
from pathlib import Path

import pytest
from flaky import flaky

from teachbooks.serve import Server

SERVE_DIR = Path(".")
WORK_DIR = Path("./.teachbooks")


@pytest.fixture
def server():
    server = Server(servedir=SERVE_DIR, workdir=WORK_DIR)
    return server

@pytest.fixture
def running_server(server):
    server.start()
    yield server
    server.stop()


@pytest.mark.parametrize(
    "port", [None, 8000]
)
def test_create(port):
    server = Server(servedir=SERVE_DIR, workdir=WORK_DIR, port=port)
    assert server.servedir == Path(".")
    assert server.workdir == Path("./.teachbooks")
    assert server.port == port
    assert server._pid == None
    assert server._statepath == Path("./.teachbooks/state.pickle")

@flaky(max_runs=10)
def test_start(running_server):
    server = running_server
    assert server.port is not None
    assert server._pid is not None
    assert server.url == f"http://localhost:{server.port}"

 
@flaky(max_runs=10)
def test_save_and_load(running_server):
    running_server._save()
    
    new_server = Server.load(WORK_DIR)
    assert new_server.servedir == running_server.servedir
    assert new_server.workdir == running_server.workdir
    assert new_server.port == running_server.port
    assert new_server._pid == running_server._pid
    assert new_server._statepath == running_server._statepath


@flaky(max_runs=10)
def test_stop(server):
    server.start()
    server.stop()
    
    assert not os.path.exists(WORK_DIR / "state.pickle")
    assert server._pid == None

@flaky(max_runs=10)
def test_multiple_start(running_server):
    pid, port = running_server._pid, running_server.port
    running_server.start()
    assert running_server._pid == pid
    assert running_server.port == port
    


import click
from pathlib import Path





@click.group()
def main():
    """TeachBooks command line tools"""
    pass


@main.command(context_settings=dict(
    ignore_unknown_options=True,
    allow_extra_args=True,
))
@click.argument("path-source", type=click.Path(exists=True, file_okay=True))
@click.option("--release", is_flag=True, help="Target release version of the book")
@click.option("--publish", is_flag=True, help="(Deprecated) Use --release instead")
@click.option("--process-only", is_flag=True, help="Only pre-process content")
@click.pass_context
def build(ctx, path_source, publish, release, process_only):
    """Pre-process book contents and run Jupyter Book build command"""
    from teachbooks.release import make_publish
    from jupyter_book.cli.main import build as jupyter_book_build

    if publish:
        click.secho("Warning: --publish is deprecated, use --release instead",
                   fg="yellow",
                   err=True)

    strategy = "release" if release or publish else "draft"
    echo_info(f"running build with strategy '{strategy}'")

    path_src_folder = Path(path_source).absolute()
    if release or publish:
        path_conf, path_toc = make_publish(path_src_folder)
    else:
        path_conf = path_src_folder / "_config.yml"
        path_toc = path_src_folder / "_toc.yml"

    if not process_only:
        # Start with base arguments
        all_args = [str(path_src_folder)]

        # Add config and toc if they exist
        if path_conf and path_conf.exists():
            all_args.extend(["--config", str(path_conf)])
        if path_toc and path_toc.exists():
            all_args.extend(["--toc", str(path_toc)])

        # Add all original args, preserving their order and values
        if ctx.args:
            all_args.extend(ctx.args)

        # Pass through all args to jupyter-book
        jupyter_book_build.main(args=all_args, standalone_mode=False)


@main.command()
@click.argument("path-source", type=click.Path(exists=True, file_okay=True))
def clean(path_source):
    """Stop teachbooks server and run Jupyter Book clean command."""
    from jupyter_book.cli.main import clean as jupyter_book_clean
    from teachbooks.serve import Server, ServerError

    workdir = Path(path_source) / ".teachbooks" / "server"

    # Check if a server is running and stop it if so
    try:
        server = Server.load(workdir)
        if server.is_running:
            echo_info("Stopping running server before cleaning...")
            server.stop()
            echo_info("Server stopped.")
    except ServerError:
        echo_info("No running server found.")

    # Now proceed with cleaning
    echo_info(f"Cleaning build artifacts in {path_source}...")
    jupyter_book_clean.main([str(path_source)])
    echo_info("Clean complete.")


@main.group(invoke_without_command=True)
@click.pass_context
def serve(ctx):
    """Start a web server to interact with the book locally"""
    from teachbooks.serve import Server

    if ctx.invoked_subcommand is None:
        # Hardcoded for now
        dir = Path("./book")
        workdir = Path("./book/.teachbooks/server")
        server = Server(servedir=dir, workdir=workdir)

        server.start(options=["--all"])
        echo_info(f"server running on {server.url}")


@serve.command()
def stop():
    """Stop the webserver"""
    from teachbooks.serve import Server
    server = Server.load(Path("./book/.teachbooks/server"))
    server.stop()
    echo_info(f"server stopped")


def echo_info(message: str) -> None:
    """Wrapper for writing to stdout"""
    prefix = click.style("TeachBooks: ", fg="cyan", bold=True)
    click.echo(prefix + message)


from dataclasses import dataclass, field

from sphinxcontrib.bibtex.style.referencing import BracketStyle
from sphinxcontrib.bibtex.style.referencing.author_year import AuthorYearReferenceStyle

def bracket_style() -> BracketStyle:
    return BracketStyle(left='(', right=')')

@dataclass
class AuthorYearRoundReferenceStyle(AuthorYearReferenceStyle):
    bracket_parenthetical: BracketStyle = field(default_factory=bracket_style)
    bracket_textual: BracketStyle = field(default_factory=bracket_style)
    bracket_author: BracketStyle = field(default_factory=bracket_style)
    bracket_label: BracketStyle = field(default_factory=bracket_style)
    bracket_year: BracketStyle = field(default_factory=bracket_style)



import re
import os
from pathlib import Path


def make_publish(sourcedir: Path) -> tuple[Path, Path]:
    """Pre-process files in a Jupyter Book directory"""
    # Make hidden directory that will contain the cleaned-up files
    workdir = sourcedir.joinpath(".teachbooks", "publish")

    if not os.path.exists(workdir):
        os.makedirs(workdir)

    for file in ["_config.yml", "_toc.yml"]:
        clean_yaml(
            sourcedir.joinpath(file),
            workdir.joinpath(file)
        )

    return workdir.joinpath("_config.yml"), workdir.joinpath("_toc.yml")


def clean_yaml(path_source: str | Path, path_output: str | Path) -> None:
    """Removes sections marked with # <START|END> REMOVE-FROM-PUBLISH or REMOVE-FROM-RELEASE from a yaml file

    Does not require a specific indentation and can be used an unlimited number of times
    in the ``*.yml`` file. Commonly applied to ``_toc.yml`` and ``_config.yml`` files of a book.

    Example:
    .. code:: python
        - file: subdirectory_1/intro_page
        sections:
        - file: subdirectory_1/sub_page_1
        # START REMOVE-FROM-PUBLISH
        - file: subdirectory_1/sub_page_2
        # END REMOVE-FROM-PUBLISH
        # START REMOVE-FROM-RELEASE
        - file: subdirectory_1/sub_page_3
        # END REMOVE-FROM-RELEASE
        - file: subdirectory_2/intro_page

    """

    with open(path_source, mode="r", encoding="utf8") as f:
        yaml_source = f.read()

    # Regex to remove both PUBLISH and RELEASE tags
    yaml_output = re.sub(
        r"# START REMOVE-FROM-(PUBLISH|RELEASE)(.|\n)*?# END REMOVE-FROM-(PUBLISH|RELEASE)",
        "",
        yaml_source
    )

    with open(path_output, mode="w", encoding="utf8") as f:
        f.write(yaml_output)



import pickle
import sys
import os
import socket
import platform
from subprocess import DEVNULL
from pathlib import Path
from typing import TypeVar
from time import sleep

import psutil

STATUS = {
    "Linux": "sleeping",
    "Darwin": "running",
    "Windows": "running"
}

Server_t = TypeVar("Server_t", bound="Server")


class ServerError(Exception):
    """Server exception class"""
    pass


class Server:
    """Class for managing a Python webserver in the background."""
    statefile = "state.pickle"

    def __init__(self, servedir: Path | str, workdir: Path | str, port: int | None = None) -> None:
        """Construct Server object

        Parameters
        ----------
        servedir : Path | str
            Directory to serve.
        workdir : Path | str
            Directory for temporary files.
        port : int | None, optional
            Port for server, by default None. If left empty, a free port will be
            chosen automatically.
        """
        # Check if workdir already contains a statefile.
        try:
            old_instance = self.load(workdir)
            self.__dict__ = old_instance.__dict__
        except ServerError:
            self.servedir = Path(servedir)
            self.workdir = Path(workdir)
            self.port = port

            self._pid = None
            self._statepath = self.workdir / self.statefile

            if not os.path.exists(self.workdir):
                os.makedirs(self.workdir)


    def start(self, options: list[str] = None) -> None:
        """Start server.

        Raises
        ------
        RuntimeError
            If server could not be started
        """
        print(f"Serving directory: {self.servedir}")  # Print the directory being served
        if not self.servedir.is_dir():
            raise NotADirectoryError(f"Directory does not exist: {self.servedir}")


        if self.port is None:
            self.port = self._find_port()
        
        if self.is_running:
            return
        else:

            base_command = [sys.executable, "-u", "-m", "http.server", str(self.port)]
            if options:
                base_command.extend(options)

            # Print the full command for verification
            print("Starting server with command:", " ".join(base_command))

            proc = psutil.Popen([sys.executable, "-u", "-m", "http.server", str(self.port)],
                                cwd=self.servedir,
                                stderr=DEVNULL,
                                stdout=DEVNULL)

            self._pid = proc.pid

            sleep(0.2)

            # Check if the subprocess is still running
            if not self.is_running:
                proc.terminate()
                raise RuntimeError("Error launching the server. Perhaps a server is already running on the selected port?")
            else:
                self._save()


    def stop(self, options: list[str] = None) -> None:
        """Stop server and clean up.
        """
        try:
            psutil.Process(pid=self._pid).terminate()
        except psutil.NoSuchProcess:
            pass

        if os.path.exists(self._statepath):
            os.remove(self._statepath)

        self._pid = None


    def _save(self) -> None:
        """Save current Server object as a pickle file.
        """
        with open(self._statepath, "wb") as f:
            pickle.dump(self, f)

    @property
    def is_running(self) -> bool:
        """Check if the current process ID is a running webserver.

        Returns
        -------
        bool
            True if the current process ID is running and is a webserver.
        """
        # Make sure the process exists
        try:
            proc = psutil.Process(pid=self._pid)
        except psutil.NoSuchProcess:
            return False
        isalive = proc.status() == STATUS[platform.system()]
        isserver = proc.cmdline()[1:] == ["-u", "-m", "http.server", str(self.port)]
        return isalive and isserver


    @property
    def url(self) -> str:
        """Get URL of running server.

        Returns
        -------
        str
            URL of the server.
        """
        return f"http://localhost:{self.port}"


    @staticmethod
    def _find_port() -> int:
        """Find open port.

        Returns
        -------
        int
            Port number.
        """
        # https://stackoverflow.com/a/1365284
        sock = socket.socket()
        sock.bind(('', 0))
        return sock.getsockname()[1]


    @classmethod
    def load(cls, workdir: Path | str) -> Server_t:
        """Construct a Server object from an existing pickle file.

        Parameters
        ----------
        workdir : Path | str
            Directory containing the pickle file.

        Returns
        -------
        Server
            Server object reconstructed from pickle file.

        Raises
        ------
        ServerError
            If pickle file cannot be found.
        """
        try:
            with open(workdir / cls.statefile, "rb") as f:
                server = pickle.load(f)
        except FileNotFoundError as exc:
            raise ServerError("Server information not found.") from exc
        return server



# -*- coding:Utf-8 -*-
from __future__ import unicode_literals

from collections import Counter
import re
import unicodedata

from pybtex.style.labels import BaseLabelStyle


_nonalnum_pattern = re.compile('[^A-Za-z0-9 \-]+', re.UNICODE)


def _strip_accents(s):
    return "".join(
        (c for c in unicodedata.normalize('NFD', s)
            if not unicodedata.combining(c)))


def _strip_nonalnum(parts):
    """Strip all non-alphanumerical characters from a list of strings.

    >>> print(_strip_nonalnum([u"ÅA. B. Testing 12+}[.@~_", u" 3%"]))
    AABTesting123
    """
    s = "".join(parts)
    return _nonalnum_pattern.sub("", _strip_accents(s))


class APALabelStyle(BaseLabelStyle):
    def format_labels(self, sorted_entries):
        labels = [self.format_label(entry) for entry in sorted_entries]
        count = Counter(labels)
        counted = Counter()
        for label in labels:
            if count[label] == 1:
                yield label
            else:
                yield label # + chr(ord('a') + counted[label])
                # counted.update([label])

    def format_label(self, entry):
        label = "Anonymous"
        if 'author' in entry.persons:
            label = self.format_author_or_editor_names(entry.persons['author'])
        elif 'editor' in entry.persons:
            label = self.format_author_or_editor_names(entry.persons['editor'])
        elif 'organization' in entry.fields:
            label = entry.fields['organization']
            if label.startswith("The "):
                label = label[4:]

        if 'year' in entry.fields:
            return "{}, {}".format(label, entry.fields['year'])
        else:
            return "{}, n.d.".format(label)

    def format_author_or_editor_names(self, persons):
        if len(persons) is 1:
            return _strip_nonalnum(persons[0].last_names)
        elif len(persons) is 2:
            return "{} & {}".format(
                _strip_nonalnum(persons[0].last_names),
                _strip_nonalnum(persons[1].last_names))
        else:
            return "{} et al.".format(
                _strip_nonalnum(persons[0].last_names))
